<?php

/**
 * Paquetes gestion WS-Security
 */
namespace WSCOMUN;

use SoapClient;
use WSSESoap;
use XMLSecurityKey;

require_once('XMLSecLibs/WSSESoap.php');

/**
 * Clase que extiende SoapClient permitiendo que consuma WS seguros
 *
 * El protocolo SOAP puede aadir soporte a la seguirdad a travs del estandar
 * {@link https://www.oasis-open.org/committees/tc_home.php?wg_abbrev=wss Web Services Security (WSS)}.
 * Actualmente el cliente php SoapClient no soporta esta extensin de forma nativa,
 * aunque con el mdulo {@link http://code.google.com/p/libwebta/source/browse/trunk/library/NET/API/AWS/ WSSESoap}
 * de la librera de utilizades {@link http://code.google.com/p/libwebta/ LibWebTA} podemos incorporarla
 *
 * @package		WSCOMUN
 * @author		David Pascual <pascual_dav@externos.gva.es>
 * @version 	$Revision$
 * @see			SoapClient
 * @since		Class disponible desde Release 1.0.0
 * Last changed: $LastChangedDate: 2018-02-21 17:36:18 +0100 (Wed, 21 Feb 2018) $
 * @author $LastChangedBy: belenguer_pab $
 *
 */
class WSSSoapClient extends SoapClient
{
	/**
	 * @var String
	 * Ruta del fichero de la clave privada
	 */
	protected $keyFile='';

	/**
	 * @var String
	 * Passphrase de la clave privada
	 */
	protected $passPhrase='';

	/**
	 * @var String
	 * Ruta del fichero del certificado
	 */
	protected $certFile='';

	/**
	 * @var $conTrazabilidadPAI boolean Indica si se incluyen o no cabeceras de trazabilidad en la invocacin
	 */
	private $conTrazabilidadPAI = true;

	/**
	 * @var $enableWSSecurity boolean Indica si se utiliza WSSecurity
	 */
	private $enableWSSecurity = true;

	/**
	 * @var $userToken string Nombre de usuario a utilizar en el UserToken
	 */
	private $userToken = '';
	
	/**
	 * @var $passUserToken string Contrasea a utilziar en el UserToken
	 */
	private $passUserToken = '';
	
	/**
	 * @var $nsCert String Nmero de serie del certificado utilizado para la invocacin de la PAI con trazabilidad
	 */
	protected $nsCert = null;

	/**
	 * @var $idApp String id de la aplicacin que invoca el WS de la PAI utilizado para la invocacin de la PAI con trazabilidad
	 */
	protected $idApp = null;

	/**
	 * @var $mtom boolean
	 * Atributo que informa si es o no es MTOM
	 */
	protected $mtom = false;


	/**
	 * Contructor. Inicializa atributos propios e invoca al constructor padre.
	 *
	 * @param  	string	$wsdl			URL asociado a los WSDLs correspondientes a cada servicio [autenticacion, autorizacion, firma]
	 * @param  	string	$vOptions		Array de opciones del cliente SOAP
	 * @access 	public
	 */
	public function __construct($wsdl, $vOptions=null)
	{

		$this->soapSent='';
		$this->keyFile='';
		$this->passPhrase='';
		$this->certFile='';

		$this->conTrazabilidadPAI = true;
		$this->enableWSSecurity = true;

		if (is_array($vOptions))
		{
			$vOptions['trace'] = true;
		}
		else
		{
			$vOptions = array('trace' => true);
		}

		return parent::__construct($wsdl, $vOptions);
	}//Fin __contruct


	/**
	 * Recibe la ruta al fichero que contiene la clave (unicamente la clave y en formato PEM)
	 *
	 * @param  	string	$pathKey	Ruta al fichero que contiene la key
	 * @access 	public
	 */
	public function setKeyFile($pathKey)
	{
		$this->keyFile = $pathKey;
	}//setKeyFile


	/**
	 * Devuelve la ruta al fichero que contiene la clave (unicamente la clave y en formato PEM)
	 *
	 * @return 	string	$pathKey	Ruta al fichero que contiene la clave
	 * @access 	public
	 */
	public function getKeyFile()
	{
		return($this->keyFile);
	}//getKeyFile


	/**
	 * Recibe la frase que permite utilizar la key
	 *
	 * @param  	string	$passphrase	Frase de paso para utilizar la clave (key)
	 * @access 	public
	 */
	public function setPassPhrase($passphrase)
	{
		$this->passPhrase = $passphrase;
	}//setPassPhrase


	/**
	 * Devuelve la frase de paso de la key o null si no tiene
	 *
	 * @return 	string	$pathCert	Frase de paso de la key
	 * @access 	public
	 */
	public function getPassPhrase()
	{
		return($this->passPhrase);
	}//getPassPhrase


	/**
	 * Recibe la ruta al fichero que contiene el cetificado (unicamente el certificado y en formato PEM)
	 *
	 * @param  	string	$pathCert	Ruta al fichero que contiene el certificado
	 * @access 	public
	 */
	public function setCertFile($pathCert)
	{
		$this->certFile = $pathCert;
	}//setCertFile

	/**
	 * Devuelve la ruta al fichero que contiene el cetificado (unicamente el certificado y en formato PEM)
	 *
	 * @return 	string	$pathCert	Ruta al fichero que contiene el certificado
	 * @access 	public
	 */
	public function getCertFile()
	{
		return($this->certFile);
	}//getCertFile


	/**
	 * Carga la informacin del certificado
	 *
	 * @param  	array	$vCertData	Array asociativo {certFile|keyFile|[passPhrase]} con ruta al certificado, ruta al fichero clave y opcionalmente frase de paso
	 * @access 	public
	 */
	public function loadCert($vCertData)
	{
		if (!is_array($vCertData))
			throw new \Exception('La informacin debe venir como array asociativo');

		if (array_key_exists('certFile', $vCertData))
		{
			$this->certFile = $vCertData['certFile'];
		}
		else
		{
			throw new \Exception('El array asociativo debe contener la ruta al certificado');
		}

		if (array_key_exists('keyFile', $vCertData))
		{
			$this->keyFile = $vCertData['keyFile'];
		}
		else
		{
			throw new \Exception('El array asociativo debe contener la ruta al fichero de clave');
		}

		if (array_key_exists('passPhrase', $vCertData))
		{
			$this->passPhrase = $vCertData['passPhrase'];
		}
	}//loadCert



	/**
	 * Devuelve si tratamos con un servicio MTOM
	 *
	 * @return 	string	$soapSent	Envo SOAP
	 * @access 	public
	 */
	public function isMTOM()
	{
		return($this->mtom);
	}//isMTOM



	// ---------------------------------------------
	// Mtodos para el uso de trazabilidad de la PAI
	// ---------------------------------------------
	/**
	 * Carga la informacin para el uso de cabeceras de trazabilidad en la PAI
	 *
	 * @param  	array	$vTrazabilidadPAI	Array asociativo {nsCert|idApp|} con num serie del certificado y el ID CATI de la aplicacin
	 * @access 	public
	 */
	public function loadTracertPAI($vTrazabilidadPAI)
	{
		if (!is_array($vTrazabilidadPAI))
		{
			throw new \Exception('Las opciones de trazabilidad deben ser un array asociativo');
		}

		if (array_key_exists('idApp', $vTrazabilidadPAI))
		{
			$this->idApp = $vTrazabilidadPAI['idApp'];
		}
		else
		{
			throw new \Exception('El array asociativo debe contener el ID de la aplicacin que consume el WS');
		}

		if ($this->enableWSSecurity)
		{
			if (!empty($this->certFile))
			{
				$this->getCertificateSerial();
			}
		}
	}//loadTracertPAI


	/**
	 * Activa el uso de cabeceras de trazabilidad en la PAI
	 *
	 * @param  	array	$vTrazabilidadPAI	Array asociativo [nsCert|idApp|conTrazabilidadPAI] si se requiera trazabilidad en la invocacin
	 * @access 	public
	 */
	public function enablePAITrace___($vTrazabilidadPAI = null)
	{
		if (empty ($vTrazabilidadPAI))
		{
			if (empty($this->idApp))
			{
				throw new \Exception(__CLASS__.":: Debe fijarse el idApp");
			}

			if ($this->enableWSSecurity())
			{
				if(empty($this->nsCert) && !empty($this->certFile))
				{
					$this->nsCert = $this->getCertificateSerial();
				}
			}
		}
		else
		{
			try
			{
				$this->loadTrazabilidad($vTrazabilidadPAI);
			}
			catch (\Exception $e)
			{
				throw $e;
			}
		}
		$this->conTrazabilidadPAI = true;
	}//enablePAITrace


	/**
	 * Desactiva el uso de cabeceras de trazabilidad en la PAI
	 * @access 	public
	 */
	public function disablePAITrace()
	{
		$this->conTrazabilidadPAI = false;
	}//disablePAITrace



	/**
	 * Activa el uso de WSSecurity
	 * @access 	public
	 */
	public function enableWSSecurity()
	{
		$this->enableWSSecurity = true;
	}//enableWSSecurity


	/**
	 * Desactiva el uso de WSSecurity
	 * @access 	public
	 */
	public function disableWSSecurity()
	{
		$this->enableWSSecurity = false;
	}//disableWSSecurity


	/**
	 * Establece los datos a utilizar en el UserToken
	 * @param  	string	$username	Nombre de usuario a emplear en el userToken
	 * @param  	string	$password	Contrasea a emplear en el userToken
	 * @access 	public
	 */
	public function setUserToken($username, $password)
	{
		$this->userToken = $username;
		$this->passUserToken = $password;
	}//setUserToken
	

	/**
	 * Devuelve el nmero de serie del certificado leyndolo del propio fichero PEM
	 *
	 * @return	string	EL nmero de serie del certificado como cadena (en hexadecimal)
	 * @access	public
	 */
	public function getCertificateSerial()
	{
		$certFile = realpath($this->certFile);
		$cert = file_get_contents($certFile);
		if ($cert === false)
		{
			throw new \Exception(__FILE__.'::'.__CLASS__." - No puedo leere el contenido de $certFile ");
		}
		$v_certData = openssl_x509_parse($cert, true);
		$this->nsCert = strtoupper(self::numberBaseConvert($v_certData['serialNumber']));
		return $this->nsCert;
	}//getCertificateSerial


	/**
	 * Recibe un string que representa un nmero, la base en la que se expresa y la base destino
	 *
	 * Utilizao debido a las limitaciones en ciertos entorno trabajando con nmeros grandes y donde puede
	 * no estar disponible la librera BC Math
	 *
	 * @param string $numstring representacin del nmero (como cadena)
	 * @param integer $frombase base origen (decimal por defecto)
	 * @param integer $tobase bse destino (hexadecimal por defecto)
	 * @return string
	 */
	private static function numberBaseConvert($numstring, $frombase=10, $tobase=16)
	{
		$chars = "0123456789abcdefghijklmnopqrstuvwxyz";
		$tostring = substr($chars, 0, $tobase);
		$length = strlen($numstring);
		$result = '';
		$number = array();
		for ($i = 0; $i < $length; $i++)
		{
			$number[$i] = strpos($chars, $numstring{$i});
		}
		do
		{
			$divide = 0;
			$newlen = 0;
			for ($i = 0; $i < $length; $i++)
			{
				$divide = $divide * $frombase + $number[$i];
				if ($divide >= $tobase)
				{
					$number[$newlen++] = (int)($divide / $tobase);
					$divide = $divide % $tobase;
				} elseif ($newlen > 0)
				{
					$number[$newlen++] = 0;
				}
			}
			$length = $newlen;
			$result = $tostring{$divide} . $result;
		} while ($newlen != 0);

		return $result;
	}


	/**
	 * Construye el token de trazabilidad: nseriecert-idApp-YYYYMMDDHHMMSSuuu.
	 *
	 * Averiguar el nro de serie de un CertDig en formato PEM: openssl x509 -in CERTIFICADO.crt -inform pem -noout -text
	 * Ejemplo: 02d8f243f4272489-WSCOMUN-20140130184955000
	 * @return string
	 */
	private function getPAITraceToken()
	{
		$t = microtime(true);
		$micro = sprintf("%06d",($t - floor($t)) * 1000000);
		$date = new \DateTime( date('Y-m-d H:i:s.'.$micro, $t));
		$token = $this->nsCert.'-'.$this->idApp.'-'.substr($date->format('YmdHisu'), 0, 17);

		return $token;
	}//getPAITraceToken




	/**
	 * Sobrecargamos el mtodo __doRequest para que se incluyan las cabeceras del protocolo SOAP WS-security. Ademas se deteca el uo de MTOM y se marca como propiedad.
	 *
	 * @param  	array	$v_wsdl	Array asociativo con los WSDLs correspondientes a cada servicio [autenticacion,autorizacion,firma]
	 * @param  	string	$v_datosCertif	Array asociativo con los datos del certificado de aplicacin necesarios para acceder. [rutaCert, rutaKey, passphrase].
	 * @return 	string	Cadena aleatorio (token) generado en el servidor
	 * @access 	public
	 */
	public function __doRequest($newRequest, $location, $action, $version, $one_way = null)
	{
		if ($this->enableWSSecurity==true)//Si trabajamos con WSSecurity
		{
			//Creamos un DOMDocument
			$doc = new \DOMDocument('1.0');

			//Cargamos el XML pasado por parmetro (WSDL)
			$doc->loadXML($newRequest);

			//Creamos una instancia de WSSESoap con el documento, sin "mustUnderstand" ni "actor"
			$objWSSE = new WSSESoap($doc, false, false);

			if (!empty($this->certFile))
			{
				//Creamos una instacian de XMLSecurityKey con el RSA SHA1 y de tipo privado
				$objKey = new XMLSecurityKey(XMLSecurityKey::RSA_SHA1, array('type'=>'private'));
				//Pasamos el valor de passphrase
				$objKey->passphrase = $this->getPassPhrase();
				//Cargamos la clave privada con la ruta del fichero, que es fichero y que no es certificado
				$objKey->loadKey($this->getKeyFile(), true);
	
				//Firmamos el mensaje
				$objWSSE->signSoapDoc($objKey);
	
				//Aadimos el BinarySecurityToken, pasamos el contenido del certificado, le decimos que es formato PEM
				$token = $objWSSE->addBinaryToken(file_get_contents($this->getCertFile()), true);
				//Aadimos el KeyInfo
				$objWSSE->attachTokentoSig($token);
			}
			elseif (!empty($this->userToken))
			{
				//Aadimos el UserToken
				$objWSSE->addUserToken($this->userToken, $this->passUserToken, $this->true);
			}
			else
			{
				// Si no se suministra certificado o UserToken se lanza una excepcion
				throw new \Exception('Al activar WSSecurity es obligatorio suministrar un certificado o los datos que componene el UserToken');
			}			
		
			//Aadimos el Timestamp
			$objWSSE->addTimestamp(3600);
		
			//Obtenemos el REQUEST modificado
			$newRequest	= $objWSSE->saveXML();
		}

		if ($this->conTrazabilidadPAI)//Si trabajamos con la trazabilidad PAI
		{
			$dom = new \DOMDocument('1.0');
			$dom->loadXML($newRequest);
			$xpath = new \DOMXpath($dom);
			$headers = $xpath->query("/*[local-name()='Envelope']/*[local-name()='Header']");

			if (($headers->length) > 0)
			{
				$header = $headers->item(0);//Localizamos la cabecera
			}
			else
			{
				$envelop = $dom->firstChild;
				$prefijo = $dom->lookupPrefix ('http://schemas.xmlsoap.org/soap/envelope/');
				$header = $dom->createElementNS ('http://schemas.xmlsoap.org/soap/envelope/',$prefijo.':Header','');
				$body = $envelop->firstChild;
				$header = $envelop->insertBefore($header, $body);
			}
			//Creamos el nodo traza
			$nodoTraza = $dom->createElementNS('http://dgti.gva.es/interoperabilidad', 'Id_trazabilidad');
			$nodoTraza->appendChild(new \DOMText( (string) $this->getPAITraceToken()));
			$header->appendChild($nodoTraza);
			unset($newRequest);$newRequest=null;gc_collect_cycles();
			$newRequest = $dom->saveXML();
		}

		//Tratamos la respuesta
		$response = parent::__doRequest($newRequest, $location, $action, $version, $one_way);

		//Si la respuesta NO contiene MTOM la devolvemos
		if (strpos($response, "Content-Type: application/xop+xml") === false)
		{
			$this->mtom = true;
		}
		return $response;
	}



}//Fin WSSSoapClient
?>